# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2019-09-13 19:50:06 UTC
from __future__ import unicode_literals
import datetime

version = "3.2.1"
githash = "ba3734222a40f4c2864c375c6639f32cd9df06cd"


major = 3
minor = 2
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 9, 13, 19, 50, 6)
debug = False

astropy_helpers_version = ""
