# Autogenerated by Astropy-affiliated package photutils's setup.py on 2019-12-09 18:21:40 UTC
import datetime

version = "0.7.2"
githash = "885aee002153269aee15c5923c662c48ad132532"


major = 0
minor = 7
bugfix = 2

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 12, 9, 18, 21, 40)
debug = False

astropy_helpers_version = "3.2.1"
